/*
 * The program is a sequence of byte-coded atoms,
 * some of them can be modified to allow optional
 * matching or repeating.
 */
#ifdef RE_DEBUG
# include <stdio.h>
# include <assert.h>
# if RE_DEBUG > 1
#  define dprintf(x) printf x
# else
#  define dprintf(x)
# endif
#else
# define assert(x)
# define dprintf(x)
#endif

#include "regexp.h"

#define STACKSIZE	100	/* max number of ?+*'s in the expression */

/*
 * Encoding of the program.
 *  The following atoms can have arguments:
 *	CHAR   [1 byte - character]
 *	CCL    [32 bytes - bitmask]
 *	STRING [1 byte - length] [2-255 bytes - string]
 *	OPEN   [2 bytes - offset to the next BAR2/CLOSE]
 *	BAR2   [2 bytes - offset to the next BAR2/CLOSE]
 *		offset for a null expression is 0.
 *	SAMEAS [1 byte - number of previously mathed substring]
 */
/* Modifiers */
#define OPTIONAL	0x80	/* The atom is optional */
#define REPEAT		0x40	/* Can be repeated */

/* Types */
#define TYPE_MASK	0x30
#define  SIMPLE		0x00	/* Simple atoms (+ code) */
#define	 OPEN		0x10	/* Openning parentesis + N */
#define  CLOSE		0x20	/* Closing parentesis + N */
#define  BAR1	 	0x30	/* Alternative separator (part 1) + N */
#define  SEQ_MASK	0x0f	/* Sequential number (N) */ 

/* Simple atoms */
#define SIMPLE_MASK	0xf
#define	 END	 	0	/* End of the program */
#define  BOL		1	/* Matches begin of line */
#define  EOL		2	/* Matches end of line */
#define  CHAR		3	/* A single character */
#define  ANY		4	/* Any character */
#define  CCL		5	/* Character class */
#define  WORDB		6	/* Beginning of a word */
#define  WORDE		7	/* End of a word */
#define  STRING		8	/* String */
#define  BAR2		9	/* Alternative separator (part 2) */
#define  CLOSE2		10	/* Close par -- assign the endp */
#define  SAMEAS		11	/* Same as previously matched substring */

#define NBPB		8			/* Bits per byte */
#define CCLSIZE		((1<<NBPB)/NBPB)	/* Size of CCL bitmask */
#define MAXSTRING	((1<<NBPB)-1)		/* Max string length */

#define SETOFF(p, v)	{ p[0] = ((v)-2); p[1] = ((v)-2) >> 8; }
#define OFF(p)		(*(unsigned char *)p | (p[1]<<8))
#define SKIPOFF(p)	(p) += 2

#define SPC	".*[\\$^+?()|"

/* Character mask -- to convert any char to unsigned */
#define UCHAR(c)	((c) & ((1<<NBPB)-1))

/* "Alphabetize" characters if character set is non-ordered */
#define U(c)	UCHAR(c)

/* Is the character a word one? */
#define ISWORD(c)	(((c) & 0200) || (c) == '_' || \
			 ('0' <= (c) && (c) <= '9') || \
			 ('A' <= (c) && (c) <= 'Z') || \
			 ('a' <= (c) && (c) <= 'z'))

/*
 * Values for r->hint:
 *	HARDCASE	- no hints, do everything in full
 *	ANCHORED	- the matching is anchored to begin of the line
 *	other		- character matching should start from
 */
#define HARDCASE	(-1)
#define ANCHORED	(-2)

#ifdef RE_DEBUG
/*
 * Dump CCL value in some readable form
 */
static char *re_ccldump(ep)
	register const char *ep;
{
	static char buf[512];	/* enough for anything */
	register char *p;
	char neg = 0;
	int c, cb;

	p = buf;
	if( ep[0] & 1 ) {
		neg = ~0;
		*p++ = '^';
	}
	cb = -1;
	for( c = 1 ; c <= (1<<NBPB) ; c++ ) {
		if( c < (1<<NBPB) &&
		    (ep[c / NBPB] ^ neg) & (1<<(c % NBPB)) ) {
			if( cb < 0 )
				cb = c;
		} else if( cb > 0 ) {
		printc:
			if( cb < 040 || cb > 0176 ) {
				*p++ = '\\';
				*p++ = '0' + (cb >> 6);
				*p++ = '0' + ((cb >> 3) & 07);
				*p++ = '0' + (cb & 07);
			} else
				*p++ = cb;
			if( c > cb+1 ) {
				if( c > cb+2 )
					*p++ = '-';
				cb = c-1;
				goto printc;
			}
			cb = -1;
		}
	}
	*p = 0;
	return buf;
}
				
/*
 * Print the regular expression in some readable form.
 */
void regprint(r)
	const regexp *r;
{
	register const char *p = r->pgm;
	register char c;
	const char *s, *fmt, *arg;
	int len;
	int notfirst = 0;

	while( (c = *p++) != END ) {

		switch( c & (OPTIONAL|REPEAT) ) {
		case 0:	              s = ""; break;
		case OPTIONAL:        s = "?"; break;
		case REPEAT:          s = "+"; break;
		case OPTIONAL|REPEAT: s = "*"; break;
		}

		switch( c & TYPE_MASK ) {
		case SIMPLE:
			switch( c & SIMPLE_MASK ) {
			case BOL:
				fmt = "^"; 
				break;
			case EOL:
				fmt = "$";
				break;
			case CHAR:
				fmt = "'%c'";
				len = *p++;
				break;
			case ANY:
				fmt = ".";
				break;
			case CCL:
				fmt = "[%*s]";
				len = 0;
				arg = re_ccldump(p);
				p += CCLSIZE;
				break;
			case WORDB:
				fmt = "<";
				break;
			case WORDE:
				fmt = ">";
				break;
			case STRING:
				fmt = "\"%.*s\"";
				len = *p++ & 0xff;
				arg = p;
				p += len;
				break;
			case BAR2:
				fmt = "|->%d";
				len = OFF(p);
				SKIPOFF(p);
				break;
			case CLOSE2:
				fmt = "))";
				break;
			case SAMEAS:
				fmt = "\\%d";
				len = *p++;
				break;
			}
			break;

		case OPEN:
			fmt = "(%d->%d";
			len = c & SEQ_MASK;
			arg = (char *)OFF(p);
			SKIPOFF(p);
			break;

		case CLOSE:
			fmt = "%d)";
			len = c & SEQ_MASK;
			break;

		case BAR1:
			fmt = "|%d";
			len = c & SEQ_MASK;
			break;
		}
		if( notfirst++ )
			printf(" ");
		printf(fmt, len, arg);
		printf(s);
	}
	printf("\n");
	if( r->hint == HARDCASE )
		;
	else if( r->hint == ANCHORED )
		printf("hint: anchored\n");
	else
		printf("hint: '%c'\n", r->hint);
}
#endif /* RE_DEBUG */

static char *re_errmsg;		/* error message */

/*
 * Calculate the size of memory required for storing
 * the expression and check the syntax.
 */
static int re_size(exp)
	register const char *exp;
{
	int plevel = 0;		/* nesting of pars */
	int pno = 0;		/* number of ('s */
	int size = 5;		/* size of an empty re */
	register int c; 	/* current scanned character */
	int prev;		/* previous atom */
	char pars[NSUBEXP];	/* stack of pars */
	char closed[NSUBEXP];	/* () pair is complete */

	prev = 0;
	bzero(closed, sizeof closed);
	while( (c = *exp++) != 0 ) {
		switch( c ) {
		case '\\':
			switch( c = *exp++ ) {
			case '>':
			case '<':
				size++;
				prev = 0;
				continue;

			case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				c -= '0';
				if( closed[c] ) {
					size += 2;
					prev = 0;
					continue;
				}
				/*FALL THROUGH*/

			case '0':
				re_errmsg = "illegal substring number";
				return 0;

			case 0:
				re_errmsg = "\\ at the end";
				return 0;
			}
			/*FALL THROUGH*/

		default:
		self_matching:
			c = 0;
			exp--;
		nextc:
			do {
				exp++, c++;
			} while( *exp != 0 && !index(SPC, *exp) );
			if( *exp == '\\' ) {
				if( *++exp == 0 ) {
					re_errmsg = "\\ at the end";
					return 0;
				}
				if( *exp != '<' && *exp != '>' &&
			   	    (*exp < '0' || *exp > '9') )
					goto nextc;
				exp--;	/* will scan \ again */
			}
			while( c > 255 ) {
				size += 257;
				c -= 255;
			}
			prev = c;
			size += (c == 1) ? 2 		/* CHAR */
			 		 : c + 2;	/* STRING */
			break;

		case '+':
		case '*':
			if( prev == -1 )
				size++;		/* CLOSE2 added */
		case '?':
			if( prev == 0 ) {
				re_errmsg = "illegal *?+?";
				return 0;
			} else if( prev > 2 )	/* Split STRING */
				size++;
			prev = 0;
			break;

		case '[':
			if( *exp == '^' )
				exp++;
			if( *exp == ']' )
				exp++;
			while( *exp != 0 && *exp != ']' )
				exp++;
			if( *exp++ == 0 ) {
				re_errmsg = "missing ]";
				return 0;
			}
			prev = 1;
			size += 1 + CCLSIZE;
			break;

		case '.':
			prev = 1;
			size++;
			break;

		case '(':
			prev = 0;
			size += 3;
			pars[plevel++] = ++pno;
			if( pno >= NSUBEXP ) {
				re_errmsg = "too many ('s";
				return 0;
			}
			break;

		case ')':
			prev = -1;
			size++;
			if( --plevel < 0 ) {
				re_errmsg = "unbalanced ()s";
				return 0;
			}
			closed[pars[plevel]] = 1;
			break;

		case '|':
			size += 4;
			prev = 0;
			break;

		case '$':
		case '^':
			prev = 0;
			size++;
			break;
		}
	}
	if( plevel > 0 ) {
		re_errmsg = "missing )";
		return 0;
	}
	return size;
}

static int re_parno;		/* number of the current () pair */
static char *re_ep;		/* current compiled expression pointer */

/*
 * Compile the regular expression
 * This function relies on error checks performed
 * earlier by re_size.
 */
static const char *re_docompile(exp, r)
	register const char *exp;
	regexp *r;
{
	register int c; 	/* current scanned character */
	int neg, xc;		/* negation & top of the region */
	char *prevp;		/* ptr to the previous atom */

	prevp = 0;
	for(;;) {
		switch( c = *exp++ ) {
		case '\\':
			switch( c = *exp++ ) {
			case '<':
				*re_ep++ = WORDB;
				prevp = 0;
				continue;
			case '>':
				*re_ep++ = WORDE;
				prevp = 0;
				continue;
			case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				prevp = 0;
				*re_ep++ = SAMEAS;
				*re_ep++ = c - '0';
				continue;
			}
			assert(c != 0);
			/*FALL THROUGH*/

		default:
			prevp = re_ep;
			re_ep += 2;
			exp--;
			c = 0;
		nextc:
			do {
				*re_ep++ = *exp++;
				if( ++c > 255 ) {
					*prevp = STRING;
					prevp[1] = 255;
					prevp = re_ep - 1;
					re_ep++;	/* skip length */
					*re_ep++ = *prevp; /* save char */
					c = 1;
				}
			} while( *exp != 0 && !index(SPC, *exp) );
			if( *exp == '\\' ) {
				assert(exp[1] != 0);
				if( *++exp != '<' && *exp != '>' &&
			   	    (*exp < '0' || *exp > '9') )
					goto nextc;
				exp--;	/* will scan \ again */
			}
			if( c == 1 ) {
				*prevp = CHAR;
				prevp[1] = *--re_ep;
			} else {
				*prevp = STRING;
				prevp[1] = c;
			}
			break;

		case '+':
			c = REPEAT;
			goto mod;

		case '*':
			c = REPEAT | OPTIONAL;
			goto mod;

		case '?':
			c = OPTIONAL;
		mod:
			assert(prevp != 0);
			/* Shorten the string... */
			if( *prevp == STRING ) {
				if( prevp[1] == 2 ) {
					*prevp = CHAR;
					prevp[1] = prevp[2];
					prevp += 2;
				} else {
					prevp[1]--;
					prevp = re_ep - 1;
					*re_ep++ = *prevp;
				}
				*prevp = CHAR;
			} else if( (*prevp & TYPE_MASK) == CLOSE && 
				   (c & REPEAT) )
				*re_ep++ = CLOSE2;
			*prevp |= c;
			prevp = 0;
			break;

		case '[':
			prevp = re_ep;
			*re_ep++ = CCL;
			neg = 0;
			if( *exp == '^' ) {
				neg++;
				exp++;
			}
			do {
				assert(*exp != 0);
				c = U(*exp++);
				re_ep[c / NBPB] |= 1 << (c % NBPB);
				if( *exp == '-' && exp[1] != ']' ) {
					xc = U(exp[1]);
					exp += 2;
					while( ++c <= xc )
						re_ep[c / NBPB] |= 1 << (c % NBPB);
				}
			} while( *exp != ']' );
			exp++;
			if( neg ) {
				for( c = 0 ; c < CCLSIZE ; c++ )
					re_ep[c] ^= ~0;
			}
			re_ep += CCLSIZE;
			break;

		case '.':
			prevp = re_ep;
			*re_ep++ = ANY;
			break;

		case '(':
			c = ++re_parno;
			*re_ep++ = OPEN | c;
			r->openpar[c] = re_ep;
			for(;;) {
				prevp = re_ep;
				SKIPOFF(re_ep);
				exp = re_docompile(exp, r);
				if( *exp == ')' ) 
					break;
				assert(*exp == '|');
				exp++;
				*re_ep++ = BAR1 | c;
				SETOFF(prevp, re_ep-prevp);
				*re_ep++ = BAR2;
			}
			SETOFF(prevp, re_ep-prevp);
			exp++;
			prevp = re_ep;
			r->closepar[c] = re_ep;
			*re_ep++ = CLOSE | c;
			break;

		case ')':
		case '|':
		case 0:
			return --exp;

		case '^':
			prevp = 0;
			*re_ep++ = BOL;
			break;

		case '$':
			prevp = 0;
			*re_ep++ = EOL;
			break;
		}
	}
	/*NOTREACHED*/
}

/*
 * Find if there is any hint about beginning of a line.
 */
static short re_gethint(ep)
	register const char *ep;
{
	register int off;
	register short hint;

	for(;;) {
		switch( (off = *ep++) & TYPE_MASK ) {
		case OPEN:
			off = OFF(ep);
			SKIPOFF(ep);
			if( (hint = re_gethint(ep)) == HARDCASE )
				return HARDCASE;
			ep += off;
			while( (*ep++ & TYPE_MASK) != CLOSE ) {
				off = OFF(ep);
				SKIPOFF(ep);
				if( hint != re_gethint(ep) )
					return HARDCASE;
				ep += off;
			}
			if( ep[-1] & OPTIONAL )
				return HARDCASE;
			return hint;

		case SIMPLE:
			switch( off & SIMPLE_MASK ) {
			case WORDB:
			case CLOSE2:
				continue;

			case CHAR:
				if( off & OPTIONAL )
					return HARDCASE;
				return UCHAR(*ep);

			case STRING:
				return UCHAR(ep[1]);

			case BOL:
				return ANCHORED;
			}
		}
		return HARDCASE;
	}
	/*NOTREACHED*/
}
 
/*
 * Compile the regular expression
 */
regexp *regcomp(exp)
	const char *exp;
{
	int size;
	regexp *r;
	char *prevp;

	/*
	 * Calculate size of the regular expression
	 */
	if( (size = re_size(exp)) == 0 ) {
	error:	regerror(re_errmsg);
		return 0;
	}

	/*
	 * Allocate memory for the regular expression
	 */
	r = (regexp *) malloc(sizeof(regexp) + size);
	if( r == 0 ) {
		re_errmsg = "not enough memory for regexp";
		goto error;
	}

	/*
	 * Compile the regular expression
	 */
	re_parno = 0;
	re_ep = r->pgm;
	*re_ep++ = OPEN;
	r->openpar[0] = re_ep;
	for(;;) {
		prevp = re_ep;
		SKIPOFF(re_ep);
		exp = re_docompile(exp, r);
		if( *exp == 0 )
			break;
		assert(*exp == '|');
		exp++;
		*re_ep++ = BAR1;
		SETOFF(prevp, re_ep-prevp);
		*re_ep++ = BAR2;
	}
	SETOFF(prevp, re_ep-prevp);
	r->closepar[0] = re_ep;
	*re_ep++ = CLOSE;
	*re_ep = END;
	assert(re_ep - r->pgm == size-1);

	/*
	 * Find some properties of an expression which can help
	 * to make matching faster.
	 */
	r->hint = re_gethint(r->pgm);
	r->magic = REGEXP_MAGIC;
	return r;
}

/*
 * Match the string with regular expression
 */
int regexec(r, str)
	regexp *r;
	const char *str;
{
	return regnexec(r, str, strlen(str), 0);
}

int regnexec(r, str0, len, startat)
	regexp *r;
	const char *str0;
	int len;
	unsigned startat;
{
	register const char *xstr;
	register const char *str;
	int ulen;
	const char *strend;
	const char *ep;
	register char c;
	register int cc;
	char rpt[NSUBEXP];
	const char *startp[NSUBEXP];
	const char *openat[NSUBEXP];
	struct {
		const char *ep;		/* the next atom to start from */	
		const char *str;	/* where to start in the sorce str */
		const char *leftbnd;	/* limitation in leftward moving */
	} stack[STACKSIZE], *sp;

	/* Check that we really do have a compiled expression */
	if( r == 0 || r->magic != REGEXP_MAGIC ) {
		regerror("no regular expression");
		return 0;
	}

	bzero(r->startp, sizeof r->startp);
	r->matchstr = 0;

	strend = str0 + (len < 0 ? -len : len);
	xstr = str0 + startat;
	if( xstr > strend )
		xstr = strend;

	/*
	 * Fast scan for the first character
	 */
fastscan:
	if( r->hint >= 0 ) {
		c = r->hint;
		if( len < 0 ) {
			do {
				if( xstr < str0 )
					return 0;
				if( *xstr-- == c )
					break;
				if( xstr < str0 )
					return 0;
			} while( *xstr-- != c );
			xstr++;
		} else {
			do {
				if( xstr >= strend )
					return 0;
				if( *xstr++ == c )
					break;
				if( xstr >= strend )
					return 0;
			} while( *xstr++ != c );
			xstr--;
		}
	}

	/*
	 * Do actual matching starting from str until strend
	 */
	ep = r->pgm;
	sp = stack;
	str = xstr;
tryagain:
	while( (c = *ep++) != END ) {
		switch( c & TYPE_MASK ) {
		case OPEN:
			assert(!(c & (OPTIONAL|REPEAT)));
			c &= SEQ_MASK;
			assert(c < NSUBEXP);
			dprintf(("OPEN %d at >%s<\n", c, str));
			startp[c] = (char *) str;
			rpt[c] = 0;
			cc = OFF(ep);
			SKIPOFF(ep);
			assert((ep[cc] & TYPE_MASK) == CLOSE || ep[cc] == BAR2);

			/* Fail on the next match if there is no options */
			if( (ep[cc] & (TYPE_MASK|OPTIONAL|REPEAT)) == CLOSE )
				break;
			if( sp >= &stack[STACKSIZE] ) {
				regerror("stack overflow");
				return 0;
			}
			openat[c] = str;
			sp->str = str;
			sp->leftbnd = str;	/* fail on the 2nd attempt */
			(sp++)->ep = cc + ep;
			dprintf(("push ep=%d width=0 str=>%s<\n", ep+cc - r->pgm, str));
			break;

		case CLOSE:		/* Successful match */
			cc = c & SEQ_MASK;
			assert(cc < NSUBEXP);
			dprintf(("CLOSE %d at >%s<\n", cc, str));
			if( !(c & REPEAT) ) {
				r->startp[cc] = (char *) startp[cc];
				r->endp[cc] = (char *) str;
				break;
			}

			/* if no characters were matched, terminate */
			if( openat[cc] == str &&
			    (!(c & OPTIONAL) || rpt[cc]) )
				goto fail;

			/* if failed, continue at CLOSE2 */
			openat[cc] = str;
			sp[-1].ep = ep;
			sp[-1].str = str;
			sp[-1].leftbnd = str;
			dprintf(("adjust ep=%d width=%d str=>%s<\n", ep - r->pgm, str - sp[-1].leftbnd, str));
			rpt[cc] = 1;

			/* Go to the opening par */
			ep = r->openpar[cc];
			cc = OFF(ep);
			SKIPOFF(ep);
			assert((ep[cc] & TYPE_MASK) == CLOSE || ep[cc] == BAR2);

			/* Fail on the next match if there is no options */
			if( ep[cc] != BAR2 )
				break;

			if( sp >= &stack[STACKSIZE] ) {
				regerror("stack overflow");
				return 0;
			}
			sp->str = str;
			sp->leftbnd = str;	/* fail on the 2nd attempt */
			(sp++)->ep = cc + ep;
			dprintf(("push ep=%d width=0 str=>%s<\n", ep+cc - r->pgm, str));
			break;

		case BAR1:
			/* Successful match -- go to close */
			assert(!(c & (OPTIONAL|REPEAT)));
			c &= SEQ_MASK;
			assert(c < NSUBEXP);
			ep = r->closepar[c];
			break;

		case SIMPLE:
			/*
			 * Repeating? Find the longest satisfying
			 * substring and record the left bondary
			 * in stack.
			 */
			if( c & REPEAT ) {
				if( sp >= &stack[STACKSIZE] ) {
					regerror("stack overflow");
					return 0;
				}
				sp->leftbnd = str;
				if( !(c & OPTIONAL) ) {
					dprintf(("OPT "));
					sp->leftbnd++;
				}
				switch( c & SIMPLE_MASK ) {
				case CHAR:
					cc = *ep++;
					dprintf(("REP CHAR %c\n", cc));
					while( str < strend && *str == cc )
						str++;
					break;

				case ANY:
					dprintf(("REP ANY\n"));
					str = strend;
					break;

				case CCL:
					dprintf(("REP CCL [%s]\n", re_ccldump(ep)));
					while( str < strend ) {
						cc = UCHAR(*str);
						if( !(ep[cc/NBPB] & (1<<(cc%NBPB))) )
							break;
						str++;
					}
					ep += CCLSIZE;
					break;

				default:
					assert(0);
				}
				if( str < sp->leftbnd )
					goto fail; 
				sp->str = str - 1;
				(sp++)->ep = ep;
				dprintf(("push ep=%d width=%d str=>%s<\n",
					ep - r->pgm, str - sp[-1].leftbnd, str-1));
				break;
			}

			/*
			 * Optional matching -- if doesn't match as is,
			 * then match to an empty string an proceed
			 * as usually (no need to touch stack).
			 * If matches, push an entry into backtracking
			 * stack and go ahead with longer (non-empty)
			 * match.
			 */
			if( c & OPTIONAL ) {
				switch( c & SIMPLE_MASK ) {
				case CHAR:
					dprintf(("OPT CHAR %c\n", *ep));
					cc = *ep++;
					if( str >= strend || *str != cc )
						continue;
					break;

				case ANY:
					dprintf(("OPT ANY\n"));
					if( str >= strend )
						continue;
					break;

				case CCL:
					dprintf(("OPT CCL [%s]\n", re_ccldump(ep)));
					if( str >= strend ) {
						ep += CCLSIZE;
						continue;
					}
					cc = UCHAR(*str);
					if( !(ep[cc/NBPB] & (1<<(cc%NBPB))) ) {
						ep += CCLSIZE;
						continue;
					}
					ep += CCLSIZE;
					break;

				default:
					assert(0);
				}

				/* the next try will be null substring */
				if( sp >= &stack[STACKSIZE] ) {
					regerror("stack overflow");
					return 0;
				}
				sp->leftbnd = str;
				sp->str = str++;
				(sp++)->ep = ep;
				dprintf(("push ep=%d width=0 str=>%s<\n", ep - r->pgm, str));
				break;
			}

			/*
			 * No options -- simple case
			 */
			switch( c & SIMPLE_MASK ) {
			case BOL:
				assert(!(c & (OPTIONAL|REPEAT)));
				dprintf(("BOL\n"));
				if( str != str0 )
					goto fail;
				break;

			case EOL:
				assert(!(c & (OPTIONAL|REPEAT)));
				dprintf(("EOL\n"));
				if( str != strend )
					goto fail;
				break;

			case CHAR:
				dprintf(("CHAR %c\n", *ep));
				if( str >= strend || *ep++ != *str )
					goto fail;
				str++;
				break;

			case ANY:
				dprintf(("ANY\n"));
				if( str >= strend )
					goto fail;
				str++;
				break;

			case CCL:
				dprintf(("CCL [%s]\n", re_ccldump(ep)));
				if( str >= strend )
					goto fail;
				cc = UCHAR(*str);
				if( !(ep[cc/NBPB] & (1<<(cc%NBPB))) )
					goto fail;
				str++;
				ep += CCLSIZE;
				break;

			case WORDB:
				assert(!(c & (OPTIONAL|REPEAT)));
				dprintf(("WORDB\n"));
				if( str > str0 && ISWORD(str[-1]) )
					goto fail;
				break;

			case WORDE:
				assert(!(c & (OPTIONAL|REPEAT)));
				dprintf(("WORDE\n"));
				if( str < strend && ISWORD(*str) )
					goto fail;
				break;

			case STRING:
				assert(!(c & (OPTIONAL|REPEAT)));
				cc = UCHAR(*ep++);	/* length */
				dprintf(("STRING \"%.*s\"\n", cc, ep));
				if( str + cc > strend || *str != *ep ||
				    memcmp(str, ep, cc) )
					goto fail;
				ep += cc;
				str += cc;
				break;

			case BAR2:
				assert(!(c & (OPTIONAL|REPEAT)));
				dprintf(("BAR2\n"));
				cc = OFF(ep);
				SKIPOFF(ep);
				assert((ep[cc] & TYPE_MASK) == CLOSE || ep[cc] == BAR2);

				/* Fail on the next match if there is no options */
				if( (ep[cc] & (TYPE_MASK|OPTIONAL|REPEAT)) == CLOSE )
					break;
				sp[-1].str = str;	/* restore str */
				sp[-1].ep = cc + ep;
				dprintf(("adjust ep=%d width=0 str=>%s<\n", ep + cc - r->pgm, str));
				break;

			case CLOSE2:
				assert((ep[-2] & TYPE_MASK) == CLOSE);
				cc = ep[-2] & SEQ_MASK;
				assert(cc < NSUBEXP);
				dprintf(("CLOSE2 %d\n", cc));
				r->startp[cc] = (char *) startp[cc];
				r->endp[cc] = (char *) str;
				break;

			case SAMEAS:
				dprintf(("SAMEAS %d\n", *ep));
				c = *ep++;
				assert(c > 0 && c < NSUBEXP);
				if( r->startp[c] == 0 )
					goto fail;
				cc = r->endp[c] - r->startp[c];
				if( cc == 0 )
					break;
				if( str + cc > strend ||
				    memcmp(str, r->startp[c], cc) )
					goto fail;
				str += cc;
				break;

			default:
				assert(0);
			}
		}
	}
	/* Matched! */
	r->matchstr = str0;
	r->matchstrlen = len;
	return 1;

fail:	/*
	 * Pop up the backtracking stack
	 */
	dprintf(("fail! -- "));
	while( sp > stack ) {
		ep = sp[-1].ep;
		str = sp[-1].str;
		if( str >= sp[-1].leftbnd ) {
			dprintf(("again ep=%d new width=%d str=>%s<\n",
				ep - r->pgm, str - sp[-1].leftbnd - 1, str));
			sp[-1].str--;	/* for one-char atoms */
			goto tryagain;
		}
		dprintf(("pop "));
		sp--;
	}
	dprintf(("-- stack empty\n"));

	/*
	 * Matching (globally) failed -- if it was an
	 * unanchored match advance pointer and try again.
	 */
	if( r->hint == ANCHORED )
		return 0;
	if( len < 0 ) {
		if( --xstr >= str0 )
			goto fastscan;
	} else {
		if( ++xstr <= strend )
		goto fastscan;
	}
	return 0;
}
